#!/usr/bin/python
# -*- coding: utf-8 -*-

# mapWOC
#
# Description: A MasterCtrl with twisted.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from ConfigParser import RawConfigParser, NoSectionError
from time import sleep
import json
import logging
import re

from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor, task

from mapwoc.lib.toolutil import MapwocOptionParser
from mapwoc.lib.protocolutil import json_trsuccess, json_trerror
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook, mapwoc_dejson

from mapwoc.masterctrl.master import Master

class Requests:
    """ Request Type 'enum'."""
    NODESTATES = 1001
    STATE = 1002
    MESSAGES = 1003
    HCS = 1005
    HC = 1006
    START_SCAN = 1007
    STOP_SCAN = 1008
    NODE = 1009
    VERSION = 1010
    CHECK_SCAN = 1011

cmd_request_map = {'state': Requests.STATE,
                   'nodestates': Requests.NODESTATES,
                   'messages':Requests.MESSAGES,
                   'hcs':Requests.HCS,
                   'hc':Requests.HC,
                   'startscan':Requests.START_SCAN,
                   'stopscan':Requests.STOP_SCAN,
                   'version':Requests.VERSION,
                   'checkscan':Requests.CHECK_SCAN,
                   'node':Requests.NODE
                   }

def which_cmd(text):
    """Returns the Request field or None."""
    
    simple_cmd = cmd_request_map.get(text)

    if simple_cmd is None:
        # Try to parse more complicated command
        tokens = text.split(" ")
        if len(tokens) == 2 and tokens[0] == 'hc':
            return Requests.HC,[tokens[1]]
        elif len(tokens) == 2 and re.match('\d', tokens[1]) and tokens[0] == 'messages':
            return Requests.MESSAGES,[int(tokens[1])]
        elif len(tokens) == 2 and tokens[0] == 'node':
            return Requests.NODE,[tokens[1]]
        else:
            return None,[]
    else:
        return simple_cmd,[]

# TODO document protocol.
class MasterRequest(LineReceiver):

    def connectionMade(self):
        """Inits empty buffer and command."""

        self.read_data = ''
        self.cmd = None

    def lineReceived(self, line):
        """Reacts on a received line."""

        self.read_data += line

        if not self.cmd:
            self.cmd,args =  which_cmd(self.read_data.strip())
            if self.cmd is Requests.NODESTATES:
                answer = self.factory.master.query_nodes_state()
                self.__answer_json(answer)
            elif self. cmd is Requests.STATE:
                answer = self.factory.master.state() 
                self.__answer_json(answer)
            elif self.cmd is Requests.MESSAGES:
                if len(args) == 0:
                    args.append(5)
                answer = self.factory.master.get_last_msgs(args[0])
                self.__answer_json(answer)
            elif self.cmd is Requests.HCS:
                answer = self.factory.master.get_hcsid_on_nodes()
                self.__answer_json(answer)
            elif self.cmd is Requests.HC:
                answer = self.factory.master.get_hc(args[0])
                self.__answer_json(answer)
            elif self.cmd is Requests.START_SCAN:
                # Read the scan configuration object.
                self.read_data = ''
            elif self.cmd is Requests.STOP_SCAN:
                self.factory.master.stop_scan()
                self.__answer(json_trsuccess())
            elif self.cmd is Requests.VERSION:
                self.__answer(self.factory.master.VERSION)
            elif self.cmd is Requests.CHECK_SCAN:
                self.read_data = ''
            elif self.cmd is Requests.NODE:
                answer = self.factory.master.get_node_by_name(args[0])
                self.__answer_json(answer)
            else:
                print "Not (yet) a command: _%s_" % self.read_data
                self.__answer('')
        else:
            #TODO its dangerous to assume no newlines are in e.g. json

            if self.cmd is Requests.START_SCAN:
                scanconfig = json.loads(self.read_data, object_hook=mapwoc_from_json_hook)
                if self.factory.master.start_scan(scanconfig):
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(333, "Did not start scan"))
            elif self.cmd is Requests.CHECK_SCAN:
                scanconfig = mapwoc_dejson(self.read_data)
                (ok,node,msg) = self.factory.master.check_scan(scanconfig)
                if ok:
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(444, "Node %s can not start scan: %s" % (node,msg)))
            

    def __answer(self, answer):
        """Sends answer through the wire and closes connection afterwards."""

        self.transport.write(answer)
        self.transport.loseConnection()


    def __answer_json(self, answer):
        """Sends json encoded answer through the wire and closes the
        connection afterwards.
        """

        json_str = MapWOCJSONEncoder().encode(answer)
        self.__answer(json_str)



class MasterResponseFactory(Factory):
    """Factory producing MasterRequest Protocols."""

    protocol = MasterRequest

    def __init__(self, master):
        self.master = master

def setup_logging(quiet, verbose):
    """Setup logging."""

    mainlogger = logging.getLogger('masterctrl')
    masterlogger = logging.getLogger('master')
    serverlogger = logging.getLogger('masterctrlsrv')
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
   
    if verbose:
        min_level = logging.DEBUG
    elif quiet:
        min_level = logging.WARNING
    else:
        min_level = logging.INFO

    for loggr in (mainlogger, masterlogger, serverlogger):
        lfh = logging.FileHandler('master.log')
        lsh = logging.StreamHandler()
        lfh.setFormatter(formatter)
        lsh.setFormatter(formatter)

        loggr.addHandler(lfh)
        loggr.addHandler(lsh)
        loggr.setLevel(min_level)

def main():
    """Startup a Reactor and MasterCtrl."""

    description = ('mapwoc-master has control over one or more mapwoc-nodes.  '
                   'mapwoc-master listens on an adress for commands and '
                   'information requests (e.g. from mapwocgui).')

    version='mapwoc-master (twisted wrapper) %s' % Master.VERSION

    combiopts = [
                ('v','verbose',"Be verbose and chatty on stdout and logfile.",
                  False,MapwocOptionParser.COMBIOPT_TYPE_MAKETRUE),
                ('q','quiet',"Be quiet, only log warnings and errors.",
                  False,MapwocOptionParser.COMBIOPT_TYPE_MAKETRUE),
                ('', 'redirector_port',"Port to talk to when talking to redirector.",
                 8558,MapwocOptionParser.COMBIOPT_TYPE_INT),
                ('', 'ctrl_port',"Port to listen on for control commands.",
                 8558,MapwocOptionParser.COMBIOPT_TYPE_INT),
                ('', 'name',"Name of the master.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'address',"Interface to listen on.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'node_config_dir',"Directory in which to look for node config files.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'credentials_path',"Path to credentials to use when syncing results with nodes.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'work_dir',"Directory to place working data.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'db',"Path to the database.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                
                ]
    parser = MapwocOptionParser(version, description,
                                "/etc/mapwoc/master/master.config", combiopts)

    (options, args) = parser.parse_args()
    if options.quiet and options.verbose:
        print "--verbose takes precedence over --quiet"
    
    setup_logging(options.quiet, options.verbose)
   
    # TODO db- integrity checks (no "open" scanrun).

    master = Master(options.config, options.node_config_dir,
                    options.credentials_path,
                    options.work_dir, options.db,
                    options.redirector_port)
    
    port = options.ctrl_port
    f = MasterResponseFactory(master)
    reactor.listenTCP(port, f)

    master_polls_nodes = task.LoopingCall(master.check_state_polls)
    master_polls_nodes.start(5)

    try:
        reactor.run()
    except:
        #reactor.stop()
        pass

    exit(0)

if __name__ == '__main__':
    main()

