# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Master-centric classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.


import ConfigParser
import datetime
import logging
from multiprocessing import Process, Queue
from os import listdir, path, makedirs
from Queue import Empty
import subprocess
import time
import socket
from time import sleep
import json

from sqlalchemy import *

from mapwocgui.model.meta import Session

from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.message import Message
from mapwoc.lib.mailhelper import MailDispatcher
from mapwoc.lib.hc import HC
from mapwoc.lib.node import Node
from nodeconnection import NodeConnection
from mapwoc.lib.scanconfiguration import ScanConfiguration
from mapwoc.lib.scanconfiguration import ScanRun
from mapwoc.lib.scanconfiguration import VisitorConfiguration
from mapwoc.lib.scanconfiguration import VisitorBundle

from nodepollprocess import NodePollProcess
from redirectorpollprocess import RedirectorPollProcess
from redirectorpollprocess import ask_redirector, ask_redirector_decode
from resultpollprocess import ResultPollProcess

from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook
from mapwoc.lib.listdict import listdict

logger = logging.getLogger('master')

class MasterState(object):
    IDLE,SCANNING,STOPPING,ERROR = range(1,5)
    strstate = {IDLE : 'idle',
                SCANNING : 'scanning',
                STOPPING : 'stopping',
                ERROR : 'error'}

    def __init__(self):
        """Initialiases Master State Set."""

        self.state = self.IDLE
        self.scanname = None
        self.nodestates = {}
        self.nodestates_are_valid = True
        self.redirectorstate = None

    def set(self, newstate):
        """Sets state of master."""

        self.state = newstate

    def export(self):
        """Returns dictionary of stateset as expected by mapwocgui."""

        return {'master' : self.strstate[self.state],
                'scanname' : self.scanname,
                'nodes' : self.nodestates,
                'redirector' : self.redirectorstate}

    def of_nodes_invalidate(self):
        """Invalidates the stateset of the nodes until they are updated"""
        self.nodestates_are_valid = False

    def of_nodes_isValid(self):
        """ Returns true if the nodestates are valid """
        return self.nodestates_are_valid

    def of_nodes_set(self, newstates):
        """Sets states of nodes."""
        self.nodestates = newstates
        self.nodestates_are_valid = True

    def of_redirector_set(self, newstate):
        """Sets state of redirector."""
        self.redirectorstate = newstate

class Master:
    """A Master."""

    VERSION = "0.2.0"

    def __init__(self, configfile, node_config_dir, credentials_path,
                 work_dir, db, redirector_port):
        """Init fields to empty lists and reads configuration file."""

        self.__nodes = {}  # Mapping node keys (names) to NodeConnections
        self.messages = [] # List of Messages
        self.__hcs = [] # List of HC's
        self.__scan = None # ScanRun Instance
        
        self.__read_conf(configfile)
        self.node_config_dir = node_config_dir
        self.credentials_path = credentials_path
        self.__work_dir = work_dir
        self.__db_file = db
        self.redirector_port = redirector_port
        self.__load_node_config()
        self.__sync_hcs()

        # Setup and start child processes that poll the other services.
        self.node_queue = Queue()
        self.node_poll_process = NodePollProcess(
                                           self.node_queue, self.__nodes, 5)
        self.node_poll_process.start()
        self.redirector_queue = Queue()
        self.redirector_poll_process = RedirectorPollProcess(
                                            self.redirector_queue, 5,
                                            port = self.redirector_port)
        self.redirector_poll_process.start()

        if not path.isdir(self.__work_dir):
            makedirs(self.__work_dir)

        engine = create_engine('sqlite:///%s' % self.__db_file)
        __c = engine.connect()
        Session.configure(bind=engine)
                                
        self.resultsyncpoller = ResultPollProcess(self.__work_dir,
                                                  self.__nodes, Queue(),
                                                  self.credentials_path,
                                                  self.__db_file)
        self.resultsyncpoller.start()
        # Delayed:
        #self.resultextractanalyser = None

        self.__state = MasterState()

    def __del__(self):
        """Deconstructor. Take care of child processes, open handles etc."""

        #TODO terminate and join the poll processes.
        pass

    def __update_redirector_state(self):
        """Sets state_set['redirector'] and places a Message."""
        try:
            # Get last answer to 'state' question of redirector.
            redirector_answer = self.redirector_queue.get_nowait()
            while not self.redirector_queue.empty():
                redirector_answer = self.redirector_queue.get_nowait()

            self.__state.of_redirector_set(redirector_answer)
        except Empty:
            logger.debug("No update from redirector state poll.")

    def __update_node_state(self):
        """Sets state_set['nodes']."""
        try:
           node_answer = self.node_queue.get_nowait()
           while not self.node_queue.empty():
               node_answer = self.node_queue.get_nowait()
           self.__state.of_nodes_set(node_answer)
        except Empty:
            logger.debug("No update from nodes state poll yet.")

    def check_state_polls(self):
        """
        Checks the Queues which are communication channels to the processes
        that ask related services about their state.  Update state.
        """

        self.__update_redirector_state()
        self.__update_node_state()

        self.__update_state()

    def __nodes_idle(self):
        """True if all nodes state is 'idle'."""

        if not self.__state.of_nodes_isValid():
            return False

        for _node,states in self.__state.nodestates.items():
            if states[''] != 'idle':
                return False
        return True

    def __update_state(self):
        """Act upon state."""

        # TODO Handle 'error' states.
        if not self.__state.of_nodes_isValid():
            return
        # Scanning:
        if self.__scan is not None:
            if self.__state.redirectorstate.get('state') == 'done':
                if self.__nodes_idle():
                    # TODO Tell redirector to idle. Listens to what it says.
                    try:
                        ask_redirector('go_idle', 'localhost', self.redirector_port)
                    except:
                        logger.warning("Couldnt ask redirector to go idle.")
                    # Tell world that scan is done.
                    self.messages.append(Message(Message.LEVEL_FUNCTIONAL_LOG,
                                         "Scan %s done" % self.__scan.scanconfiguration.name,
                                         heading="Scan"))
                    self.__maildispatcher.send_scan_finished(
                            scanname=self.__scan.scanconfiguration.name)
                    self.__scan.stoptime = datetime.datetime.now()
                    self.__scan.status = "done"
                    Session.add(self.__scan)
                    Session.commit()
                    self.__scan = None
                    self.__state.set(MasterState.IDLE)
                else:
                    for _name,node in self.__nodes.items():
                        node.ask('stoprun')
                    self.__state.set(MasterState.STOPPING)
            elif self.__state.state == MasterState.STOPPING and self.__nodes_idle():
                # Tell world that scan is aborted.
                self.messages.append(Message(Message.LEVEL_FUNCTIONAL_LOG,
                                     "Scan %s aborted" % self.__scan.scanconfiguration.name,
                                     heading="Scan"))
                self.__scan.stoptime = datetime.datetime.now()
                self.__scan.status = "aborted"
                Session.add(self.__scan)
                Session.commit()
                self.__scan = None
                self.__state.set(MasterState.IDLE)
            elif self.__state.state != MasterState.STOPPING and not self.__nodes_idle():
                # Nodes are scanning and we are not stopping
                self.__state.set(MasterState.SCANNING)
        else:
            self.__state.set(MasterState.IDLE)
            logger.warning("Setting own state to idle.")
            # TODO The other components should idle, too!

        if self.__scan:
            self.__state.scanname = self.__scan.scanconfiguration.name
        else:
            self.__state.scanname = None

        # TODO Spell out rules. Keep in mind that state information might be relatively old.
        # If self running/stopping and nodes idle and redirector idle or done, say scan done, become idle
        # If self running and redirector done, become stopping, make nodes stopping
        # If self running, nodes stopped, redirector done, idle the redirector
        # If self stopping, nodes running, redirector not done: abort scan, become idle

    def get_node_by_name(self, name):
        """ Get a Node object by it's name """

        return self.__nodes[name]

    def get_nodes(self):
        """Accessor for nodes."""

        return self.__nodes

    def check_scan(self, scanconf):
        """
        Checks whether a scan could be started by the nodes.
        @param scanconf: The Scan Configuration to let the nodes check.
        @return Triple, first value being True if all nodes could start that
                scan.  Otherwise (False,nameofnode,messge).
        """

        nodetasks = listdict()

        # Sort single HCs into correct nodetasks pocket.
        for visitor in scanconf.visitor_configuration.visitors:
            node = self.__nodes.get(visitor.node)
            if node is None:
                logger.info("Node %s for which jobs exist not found." % \
                        visitor.node)
            else:
                nodetasks.lappend(visitor.node, visitor)

        for node, visitorlist in nodetasks.items():
            node_vconf = VisitorConfiguration()
            node_vconf.visitors = visitorlist
            node_vconf.active_window = scanconf.visitor_configuration.active_window
            n = self.__nodes.get(node)
            ok,tr = n.ask_ok_tr("checkjob\r\n%s" 
                                % MapWOCJSONEncoder().encode(node_vconf))
            if not ok:
                logger.warning("Node will not be able to do the job: %s." % tr.message)
                return (False,node,tr.message)

        return (True,None,None)
            

    def start_scan(self, scanconf):
        """
        Start a ScanRun with a given ScanConfiguration.

        @param scanconf: The ScanConfiguration to distribute to nodes and
                         Redirector.
        @return False in case of errors, True otherwise.
        """

        # TODO Check scan with nodes before starting.
        # TODO respect own state and maybe already running scan conf.

        if scanconf is None:
            logger.warning("Given no ScanConfiguration.")
            return False
        vc = scanconf.visitor_configuration
        if vc is None:
            logger.warning("Given Scan Configuration lacks visitor configuration.")
            return False
        if not vc.visitors:
            logger.warning("Given Vistor Configuration does not have Visitor entries.")
            return False

        nodetasks = listdict()
        self.__scan = ScanRun()
        self.__scan.scanconfiguration_id = scanconf.id
        self.__scan.status = "start prepared"

        Session.add(self.__scan)
        Session.commit()

        collections=""
        # Create a colon seperated list of urlcollection ids.
        for id in scanconf.urllistIDs:
            collections += "%d," % id
        # Give order to redirector.
        r_response = ''
        try:
            r_response = ask_redirector_decode('loaddb %s,%s' % (collections,
                path.abspath(self.__db_file)), 'localhost', self.redirector_port)
        except:
            logger.warning("Could not ask or decode redirectors answer"
                           " to loaddb question")
            r_response = ''

        if r_response == None:
            logger.warning(
                    "panic: Redirector did not answer to loaddb question")
        if r_response.error != 0:
            logger.warning("Redictor answered %s " % r_response.message)
            return False

        # Communicate amout of URLs to scan at once.
        try:
            r_response = ask_redirector('fastmode %d' % vc.num_urls,
                    'localhost', self.redirector_port)
        except:
            logger.warning("Could not ask redirector to go to fast mode.")
        logger.debug("Asking redirector to move to fastmode %d " % vc.num_urls)
        if r_response != 'ok':
            logger.warning(
                    "panic: Redirector did not acknowledged fastmode question")

        # Sort single HCs into correct nodetasks pocket.
        for visitor in vc.visitors:
            node = self.__nodes.get(visitor.node)
            if node is None:
                logger.info("Node %s for which jobs exist not found." % \
                        visitor.node)
            else:
                nodetasks.lappend(visitor.node, visitor)

        # Create visitorconfiguration, add to scanconf, send to node.
        for node, visitorlist in nodetasks.items():
            node_vconf = VisitorConfiguration()
            node_vconf.visitors = visitorlist
            node_vconf.active_window = vc.active_window
            n = self.__nodes.get(node)
            job_description = scanconf
            job_description.visitor_configuration = node_vconf
            question = "startrun\r\n%d\r\n%s" % ( self.__scan.id,
                                  MapWOCJSONEncoder().encode(job_description))
            response = n.ask_decode(question)
            if response.error != 0:
                logger.warning("Node answered: %s" % response.message)
                return False
            else:
                self.__state.of_nodes_invalidate()
                logger.debug("Node accepted to start scan.")

        logger.info("Starting scan.")
        self.messages += [Message(Message.LEVEL_FUNCTIONAL_LOG,
                                  "Starting Scan %s" % scanconf.name,
                                  heading="Scan")]
        # Set State.
        self.__state.set(MasterState.SCANNING)
        self.__state.scanname = scanconf.name
        return True

    def stop_scan(self, graceful=False):
        """Stops a running scan."""

        if not self.__scan:
            logger.warning("No Scan found to stop.")
            return False
       
        # TODO Tell all nodes to stop.
        for v in self.__scan.scanconfiguration.visitor_configuration.visitors:
            response = self.__nodes[v.node].ask('stoprun')

        try:
            ask_redirector('go_idle', 'localhost', self.redirector_port)
        except:
            logger.warning("Couldnt ask redirector to go idle.")

        # Collect the answers from the stop questions, remove the scan config if
        # successfull. (TODO)

        self.__state.set(MasterState.STOPPING)
        return True

    def state(self):
        """Returns string representation of state."""

        return self.__state.export()

    def __sync_hcs(self):
        """Syncs HCs with configured nodes.
           Calling might block for a longer time.
        """
        for key, node in self.__nodes.items():
            logger.debug("Getting hc's from: %s" % key)
            hcs = node.ask_decode('hcs')
            if not hcs:
                logger.warning("Node %s reports no HC's." % key)
                continue
            for hc in hcs:
                # Check if hcid is already known
                for known in self.__hcs:
                    if known.hcsid == hc.hcsid:
                        logger.debug("Adding known hc: %s " % hc.hcsid)
                        known.on_nodes.append(key)
                        break
                else:
                    logger.debug("Adding unknown hc: %s " % hc.hcsid)
                    hc.on_nodes = [key]
                    self.__hcs.append( hc )


    def sync_results(self):
        """Syncs data from nodes by invoking an rsync-script (blocks)."""

        # poll the queues?
        pass

    def __read_conf(self, config_file):
        """Reads in configuration from file."""

        config = ConfigParser.RawConfigParser ()
        config.read(config_file)
        try:
            self.__maildispatcher = MailDispatcher(
                    config.get('notifications', 'smtp_server'),
                    config.get('notifications', 'smtp_port'),
                    config.get('notifications', 'admin_mail'),
                    config.get('notifications', 'sender_mail'))
        except:
            logger.info("No valid mail configuration found. "
                        "Disabling Notifications.")
            self.__maildispatcher = MailDispatcher(
                    None, None, None, None)

    def __load_node_config(self):
        """Load the node config files from node config directory."""
        
        logger.info("Loading nodes from %s." % self.node_config_dir)

        confdir = self.node_config_dir
        # Expand and normalize all paths.
        abs_file_paths = [ path.normpath(path.join(confdir, file_rel))
                           for file_rel in listdir(confdir) ]
        # Just keep the ones ending in .conf .
        abs_conf_files = [ file_abs for file_abs in abs_file_paths
                           if file_abs.endswith(".config") ]

        # Parse the remaining ones and add to list of nodes.
        for node_conf_file in abs_conf_files:
            node = NodeConnection(node_conf_file)
            self.__nodes[node.key] = node
            node.data_path = node.ask('resultdir')
            if not node.data_path.startswith('/'):
                logger.error("Could not get resultsdir from node %s " % node.key)
                node.data_path = None

        logger.info("Loaded %d nodes." % len(self.__nodes))


    # TODO Obsolete, remove together with masterctrl.py .
    def get_node_confs(self):
        """Return the config file for all nodes as string."""

        allnodesconf = ""

        for rel_file in listdir(self.node_config_dir):
            abs_file = path.join(self.node_config_dir, rel_file)
            if path.isdir(abs_file) == False:
                allnodesconf += "\n".join(open(abs_file,'r').readlines())

        return allnodesconf


    # Obsolete, remove together with masterctrl.py .
    def get_node_conf(self, nodekey):
        """Return the config file for a node as string."""

        return "\n".join(open(path.join(self.node_config_dir,
                                        nodekey + ".config"),'r'))


    def get_hcsid_on_nodes(self):

        answer = {}
        for hc in self.__hcs:
            for node in hc.on_nodes:
                if answer.get(hc.hcsid):
                    answer[hc.hcsid].append(node)
                else:
                    answer[hc.hcsid] = [node]
        return answer

    def get_hc(self, hcsid):

        for hc in self.__hcs:
            if hc.hcsid == hcsid:
                return hc
        return None


    def get_last_msgs(self, num=1):
        """Return the num last messages in a list.
           The first item in list is last message.
        """

        msgs = self.messages[-num:]
        msgs.reverse()
        return msgs

