# -*- coding: utf-8 -*-

# mapWOC
#
# Description: The Server to receive master-commands.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import SocketServer
import re
import logging
import json

from mapwoc.lib.message import Message
from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook as json_decode
from mapwoc.lib.protocolutil import json_trerror, json_trsuccess, json_tr_unknown_command

logger = logging.getLogger('masterctrlsrv')

class MasterCtrlServer(SocketServer.TCPServer):
    """A simple TCPServer with a MasterCtrlServerReqHandler."""

    def __init__ (self, address, master):
        """Initializes a MasterCtrlServer with a Master instance."""
        SocketServer.TCPServer.__init__(self, address,
                                        MasterCtrlServerReqHandler)
        self.master = master
        self.allow_reuse_address = True


class MasterCtrlServerReqHandler(SocketServer.BaseRequestHandler):
    """Request Handler of a MasterCtrlServer."""


    def handle(self):
        """Handle a masterctrl request."""

        logger.debug("Incoming MasterCtrl request.")
        question = self.request.recv(2048).strip()
        answer = ""

        # Query State
        if question == 'state':
            answer = self.server.master.state()

        # Start a Scan.
        elif question.startswith('startscan'):
            scan = json.loads(question[10:], object_hook=json_decode)
            if self.server.master.start_scan(scan):
                answer = json_trsuccess()
            else:
                answer = json_trerror(1, "Could not start scan.")

        # Stop a Scan.
        elif question.startswith('stopscan'):
            if self.server.master.stop_scan():
                answer = json_trsuccess()
            else:
                answer = json_trerror(1, "Could not stop scan.")


        # Get (4) last Messages.
        elif question == 'messages':
            #answer = str(self.server.master.messages)
            last_msgs = self.server.master.messages[-4:]
            last_msgs.reverse()
            logger.debug("have %d messages." % len (self.server.master.messages))
            answer = MapWOCJSONEncoder().encode(last_msgs)

        elif question == 'synchcs':
            answer = self.server.master.sync()
        elif question == 'nodes':
            #self.server.master.sync()
            for key, node in self.server.master.get_nodes().items():
                answer += "%s (%s)\n" % (node.name, node.ip)
        elif question == 'nodestates':
            answer = self.server.master.query_nodes_state()
        elif question.startswith('nodectrl '):
            #parse the node
            match_res = re.match('nodectrl (.*) starthc (.*)', question)
            if match_res is None:
                match_res = re.match('nodectrl (.*) stophc (.*)', question)
                if match_res is not None:
                    for key,n in self.server.master.get_nodes().items():
                        if n.key == match_res.group(1):
                            answer = n.ask('stopallhcs')
                            self.server.master.messages += [Message(
                                        Message.LEVEL_RESULT, "Stop all hcs.")]
                            break
                match_res = re.match('nodectrl (.*) runlist', question)
                # TODO resolve this. its rediculous
                if match_res is not None:
                    for key, node in self.server.master.get_nodes().items():
                        if node.key == match_res.group(1):
                            answer = node.ask('list run')
                            break
                else:
                    logger.warning("nodectrl ("+question+") " \
                           "command in master not parseable.")
                    answer = "invalid command"
            else:
                # TODO Resolve this ridicoulousity.
                for key,node in self.server.master.get_nodes().items():
                    if node.key == match_res.group(1):
                        answer = node.ask('starthc ' + match_res.group(2))
                        self.server.master.messages += [Message(
                                    Message.LEVEL_FUNCTIONAL_LOG,
                                    "Start hc on %s" % match_res.group(1))]
                        break
        elif question.startswith('nodestate '):
            node_id = question[10:]
            answer = self.server.master.get_node_state(node_id)
        elif question.startswith('hcs '):
            node_id = question[4:]
            self.server.master.load_node_hcs()
            for hc in self.server.master.hcs:
                if node_id in self.server.master.hcs[hc].on_nodes:
                    answer += hc + "\n"
        elif question.startswith('hc '):
            hcsid = question[3:]
            # To return values from the struct:
            #self.server.master.load_node_hcs()
            #hc = self.server.master.hcs.get(hcsid, None)
            #if hc is not None:
            #    answer = "hc " + hc.name + " description: " + hc.description
            #else:
            #    answer = "hc " + hcsid + " not found."

            # To return the configfile:
            answer = self.server.master.get_hc_conffile(hcsid)
        elif question == 'unique hcs':
            answer = "no unique hcs"
        elif question == 'hcs':
            hcs = self.server.master.load_node_hcs()
            for hcname in hcs:
                answer += "%s: %s\n" % (hcs[hcname].hcsid, hcs[hcname].name)
        elif question == 'nodeconfs':
            answer = self.server.master.get_node_confs()
        elif question.startswith('nodeconf '):
            answer = self.server.master.get_node_conf(question[9:])
        else:
            answer = json_tr_unknown_command()
            logger.warning("Unknown masterctrl command: %s." % question.strip())

        self.request.send(answer)

