# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process results
#
# Authors:
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
from datetime import datetime
import logging
import os
import shutil
import tarfile
import traceback

from mapwoc.lib.result import Result, Visit, ResultType, ResultFile
from mapwoc.lib.scanconfiguration import ScanConfiguration, ScanRun
from mapwocgui.model.meta import Session
from mapwoc.lib.url import URL

logger = logging.getLogger('master')

DEBUGANALYSER = True

class ResultAnalyser(object):
    """ Provides a base class for Result Analysers """
    # NOTE: some functions should be protected in here is that possible?

    def __init__(self, result_data, node):
        """
        Resultanalyser class to enter raw result_data into the Database as
        results.

        @param result_data qualifier of the raw data which should be parsed.
        @param node Information on which node the scan was running.
        """
        if DEBUGANALYSER:
            logger.debug("ResultAnalyser __init__ with data %s for node %s" %\
                    (result_data, node))
        self.result_data = result_data
        self.visit = Visit()
        self.visit.filepath = result_data
        self.visit.node = node

    def scan_metainformation(self, fptr):
        """
        Scan a Meta Information file.

        @param filepath File with meta information about Visit.
        """
        parser = ConfigParser.RawConfigParser()
        parser.readfp(fptr)

        self.visit.scanrun_id = parser.getint("scan", "scanrunid")
        self.visit.date = datetime.fromtimestamp(parser.getfloat("scan", "time"))

        scanrun = Session.query(ScanRun).filter_by(id=self.visit.scanrun_id).first()
        if DEBUGANALYSER:
            logger.debug("Metainformation scanrun.id %d" % self.visit.scanrun_id)
        self.visit.scanrun = scanrun

    def construct_result(self, url_string, type_name, method):
        """
        Construct a Result object.

        @param url: URL (string) of this result.
        @param type_name: Type of the result as a string.
        @param method: Method with which the result was obtained.
        @return: Result object with valid type but wihout meta information.
        """

        rtype = Session.query(ResultType).filter(
                                    ResultType.type_name == type_name).first()
        if rtype is None:
            rtype = ResultType(rtype, 0)
            Session.add(rtype)
            logger.warning("Added unknown result type %s " % type_name)
        if DEBUGANALYSER:
            logger.debug("Adding result:\nURL: %s\nType: %s\nMethod:%s" %\
                    (url_string, type_name, method))
        result = Result(url_string, result_type=rtype, method=method)

        return result

    def commit(self):
        """
        Commits the results of the analyser to the database and
        adds metadata which is only available after a full
        scan.
        """
        for result in self.visit.results:
            result.date = self.visit.date # TODO be more accurate
            Session.add(result)
            if result.result_file:
                Session.add(result.result_file)
        Session.add(self.visit)
        Session.commit()

    def is_interesting(self):
        """
        Check if the results are interesting.

        @return True if something interesting has been found.
        """
        # For now everything is interesting
        return True
        interesting = False
        for result in self.visit.results:
            if result.result_type.severity > 100:
                return True
        return False

    def archive_or_delete(self):
        """ Archives intersting results and deletes uninteresting ones. """
        if not self.is_interesting():
            os.remove(self.visit.filepath)
            self.visit.filepath = None

    def generic_analyse(self):
        """ Runs generic steps of the analysis. """
        if self.visit.results:
            self.visit.interesting = self.is_interesting()
            self.commit()
            self.archive_or_delete()
            logging.debug("Added new visit into database with %d results." %\
                    len(self.visit.results))
            return True

        logging.warning("Sucessfully scanned %s but did not get any results."%\
                self.result_data)
        return False


class TarballAnalyser(ResultAnalyser):
    """
    Class to analyse tarballs of Results provided
    by the mapwocanalyse.pl script.
    """

    def __init__(self, result_data, node):
        ResultAnalyser.__init__(self, result_data, node)

    def analyse(self):
        """
        Starts the analysis of a result.
        @return True if at least one result could be entered into
                the database.
        """
        try:
            self.__parse_tarball(self.result_data)
            return self.generic_analyse()
        except:
            traceback.print_exc()
            logging.error("Invalid tarball %s ignoring result." % \
                    self.result_data)
            Session.rollback()
            return False

    def __parse_tarball(self, resultb):
        """
        Imports a result archive into the db.

        @parameter resultb: should point to an existing tarball of results.
        """

        if DEBUGANALYSER:
            logger.debug("Entering result %s into database" % resultb)
        tgz = tarfile.open(resultb)

        if DEBUGANALYSER:
            logger.debug("Files in tarball: %s" % tgz.getnames())
        self.__scan_headers(tgz.extractfile('result/Log/header.log'))

        log302file = 'result/Log/302.log'
        if log302file in tgz.getnames():
            self.__scan_302(tgz.extractfile(log302file))

        testLog = 'result/Log/Test.log'
        if testLog in tgz.getnames():
            self.__scan_testlog(tgz.extractfile(testLog))

        metafile = 'metainfo'
        if metafile in tgz.getnames():
            self.scan_metainformation(tgz.extractfile(metafile))

        if DEBUGANALYSER:
            logger.debug("URLs: %s" % ", ".join([url.url for url in self.visit.urls]))

        for fname in tgz.getnames():  # Main loop to look for interesting stuff.
            # Delegate interesting files to subfunctions.
            # TODO find more interesting files or ignore all the packet files.
            if fname == 'result/Log/404.log': # At least one 404
                self.__scan_404(tgz.extractfile(fname))
                continue
            if fname == 'result/Log/Analyse.log':
                self.__scan_analyse(tgz.extractfile(fname))
                continue
            if fname.startswith("result/Binary"):
                if fname.startswith("result/Binary/GOOGLE"):
                    # TODO Could also be evil
                    continue
                self.__scan_binary(tgz.extractfile(fname))
                continue

    def __scan_testlog(self, log):
        """ Scan the TestLog and extract the hcsid from the dump file name """
        hc = log.readlines()[0].rsplit("/", 1)[1].strip()
        # Remove the 6 tmp chars, _ and .dump and snapshot
        self.visit.hcsid = hc[:len(hc)-21]
        logger.debug("Found Test.log extracted hcsid: %s " % self.visit.hcsid)

    def __scan_binary(self, fname):
        pass

    def __scan_analyse(self, log):
        lines = log.readlines()
        method = "mapwocanalyse.pl/Analyse.log/0.0.2"

        # TODO understand the structure of this file
        # and why there are duplicates of the same line
        # For now just add one result
        for line in lines:
            # TODO maybe we could remove duplicate lines before walking over.
            if DEBUGANALYSER:
                logging.debug("__scan_analyse: line %s " % line)
            recs = line.strip().split()
            analyse_type = recs[0]
            if analyse_type == "[GOOGLE]":
                host = recs[3]
                referer = recs[4]
            
                # When browsing manually we do not have an URL (that is in db)
                if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                    url_str = self.visit.urls[0].url
                else:
                    url_str = "manual browsing"
                result = self.construct_result(url_str, "GoogleAnalytics",
                                               method)
            
                result.text = "%s\n%s" % ( host, referer )
                self.visit.results.append(result)
            elif analyse_type == "[EXE]":
                host = recs[3]
                referer = recs[4]
                path = recs[-1]
            
                # When browsing manually we do not have an URL (that is in db)
                if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                    url_str = self.visit.urls[0].url
                else:
                    url_str = "manual browsing"

                result = self.construct_result(url_str, "WindowsExecutable",
                                               method)
            
                result.text = ("Windows PE32 found at %s."
                        " (from %s\n, Referer: %s)" % ( path, host, referer ))
                resultfile = ResultFile()
                resultfile.filepath = recs[-2]
                resultfile.sha1sum = "sha1sum to be computed"
                resultfile.ssdeep = "ssdeep sum"
                result.result_file = resultfile
                self.visit.results.append(result)
            else:
                # Assume its google analyse.
                host = recs[3]
                referer = recs[4]
            
                # When browsing manually we do not have an URL (that is in db)
                if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                    url_str = self.visit.urls[0].url
                else:
                    url_str = "manual browsing"
                result = self.construct_result(url_str, "GoogleAnalytics",
                                               method)
 
                result.text = "%s\n%s" % ( host, referer )
                self.visit.results.append(result)
                

    def __scan_headers(self, log):
        # TODO Making hard assumptions (http)
        lines = log.readlines()
        if len(lines) <= 1:
            # Empty header log, TODO does this mean url can not be resolved?
            logger.debug("Empty header log")
            return
        else:
            # fields in line: src-ip,time,dst(host),agent(crippled)
            # We assume that the redirector was queried in line 1 and
            # redirected to a Url.
            # If this is not the case something is broken.
            try:
                # TODO take this from redirection log
                #url_string = "http://%s" % lines[2].split()[2]
                #url = Session.query(URL).filter(URL.url == url_string).first() or \
                #      Session.query(URL).filter(URL.url == url_string.rstrip('/')).first()
                #self.visit.urls.append(url)
                #Session.rollback()
                for line in lines[2:]:
                    if DEBUGANALYSER:
                        logger.debug("__scan_headers: line: %s " % line)
                    recs = line.strip().split()
                    logger.debug("Making hard assumptions (http) about analysed url visit.")
                    url_string = "http://%s" % recs[2]
                    result = self.construct_result(url_string, "RegularTraffic",
                            "mapwocanalyse.pl/header.log/0.0.2")
                    result.text = '- found in Log/header.log -'
                    self.visit.results.append(result)
                    if DEBUGANALYSER:
                        logger.debug("Added RegularTraffic result")
                    #TODO its an error if there is no lines.
                    #    resulttype = ResultType('NoTraffic', 204)
                    #else:
                    #    if len(hl.readlines()) > 2:
                    #        resulttype = ResultType('RegularVisit', 200)
                    #    else:
                    #        resulttype = ResultType('InsufficientTraffic', 204)
                    # could find google analyse in Log/Analyse.log
            except:
               result = self.construct_result("n/a", "MissingTraffic",
                                              "mapwocanalyse.pl/header.log/0.0.2")
               self.visit.results.append(result)
               if DEBUGANALYSER:
                   logger.debug("Something went wrong. Added MissingTraffic result.")
                   traceback.print_exc()


    def __scan_404(self, log):
        """ Parses a 404.log and creates the relevant results. """

        for line in log.readlines():
            url_string = line.split('\t')[-1].strip()
            result = self.construct_result(url_string, "404",
                            method='mapwocanalyse.pl/404/0.0.2')
            result.text = line.strip()
            self.visit.results.append(result)
            logger.debug("Added 404 result")

    def __scan_302(self, log):
        """Parses a 302.log and registers the URLs from redirector."""

        # TODO 302s can be interesting results, too.
        for line in log.readlines():
            records = line.split()
            # TODO Remove hard-coded redirector address.
            logger.debug("302: %s" % line)
            logger.debug("302 - r3 : %s" % records[-2])

            if records[3] == "10.0.0.1":
                # Found our redirector.
                # This URL shall be in database.
                # TODO use sqlalchemys _o  .
                url = Session.query(URL).filter(URL.url == records[-2]).first() or \
                      Session.query(URL).filter(URL.url == records[-2].rstrip('/')).first()
                if url:
                    self.visit.urls.append(url)
                if DEBUGANALYSER:
                    logger.debug("Added an URL from 302.log.")

