# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Analyser-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
import logging
import subprocess
import tarfile
import time
from time import sleep
import tempfile
from shutil import copy, move

from os import remove
from os.path import basename, abspath

from mapwoc.lib.result import Result

logger = logging.getLogger("node")

class AnalysisProcess(multiprocessing.Process):

    SLEEP_STEP = 1

    def __init__(self, job_queue, result_queue, working_dir, result_dir,
            analyser = '/usr/bin/mapwocanalyse.pl'):
        """Constructor.
           job_queue Queue for (dump-)filenames to be analysed.
           result_queue Queue for results to be put when analysis finished.
           working_dir Working directory (where temp. files will be placed.)
           result_dir Directory where result tarballs will be placed.
           analyser Path to the analyser which should be used
        """

        multiprocessing.Process.__init__(self)
        self.result_queue = result_queue
        self.job_queue = job_queue
        self.working_dir = working_dir
        self.result_dir = result_dir
        self.analyser = analyser

    def run(self):
        """Invoked on start() - call.
           Polls Queue for (dump-)filenames and analyses these if possible.
        """

        while True:

            job = self.job_queue.get()

            logger.debug("Starting analysis of: %s" % job)

            # Duplicate not necessary.
            job_file = job

            tempdir = abspath(
                    tempfile.mkdtemp(dir=self.working_dir, prefix='anaylyser_'))
            # TODO When copying fails, node should go into error state.
            copy(job_file, tempdir)

            # Obtain a temporary working dir.
            cmd = [self.analyser, basename(job_file), tempdir]

            logger.debug("Launching Analysis: %s", " ".join(cmd))

            process = subprocess.Popen(cmd, close_fds=True,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            # Wait for process to finish.
            process.wait()
            if process.returncode:
                stdout_value, stderr_value = process.communicate()
                logger.debug("stdout: %s" % stdout_value)
                logger.debug("stderr: %s" % stderr_value)
                logger.debug("exit code: %d" % process.returncode)

            # Now bundle the results.
            _fh,fname = tempfile.mkstemp(dir=self.working_dir, suffix='.tar')
            tgz = tarfile.open(name=fname, mode='w')
            tgz.add(tempdir, arcname='result')
            tgz.close()

            # TODO How atomic is copying the targz?
            # (regarding rsync) would a file lock help?

            remove(job_file)
            self.result_queue.put(fname)
