#!/usr/bin/python
# -*- coding: utf-8 -*-

# mapWOC
#
# Description: A NodeCtrl with twisted.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from ConfigParser import NoSectionError
import logging

from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor, task
from time import sleep
import json

from mapwoc.lib.toolutil import MapwocOptionParser

from mapwoc.lib.protocolutil import json_trsuccess, json_trerror
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook

from mapwoc.nodectrl.node import Node

# TODO There is severe overlap between this and the master implemenation.
#      Eventually, it might be refactored/merged.

class Requests:
    """Request type 'enum'."""
    STATE = 1002
    HCS = 1003
    START_RUN = 1004
    STOP_RUN = 1005
    RESULT_DIR = 1006
    VERSION = 1010
    CHECK_RESSOURCES = 1011

cmd_request_map = {'state':Requests.STATE,
                   'hcs':Requests.HCS,
                   'startrun':Requests.START_RUN,
                   'stoprun':Requests.STOP_RUN,
                   'version':Requests.VERSION,
                   'checkjob':Requests.CHECK_RESSOURCES,
                   'resultdir':Requests.RESULT_DIR
                  }


def which_cmd(text):
    """Returns Requests from string."""

    return cmd_request_map.get(text)

class NodeRequest(LineReceiver):
    """Handle one (control or info) request."""

    def connectionMade(self):
        """Just prepare for data coming."""

        self.text = ''
        self.cmd = None
        self.nr_lines_received = 0

    def lineReceived(self, line):
        """
        Callback, received a line.
        React upon it and answer, or wait for further data (lines).
        """

        self.text += line
        self.nr_lines_received = self.nr_lines_received + 1

        if not self.cmd:
            self.cmd = which_cmd(self.text.strip())

            if self.cmd is Requests.STATE:
                answer = self.factory.node.state.export()
                self.__answer_json(answer)
            elif self.cmd is Requests.HCS:
                answer = self.factory.node.get_hcs()
                self.__answer_json(answer)
            elif self.cmd is Requests.START_RUN:
                self.text = ''
            elif self.cmd is Requests.CHECK_RESSOURCES:
                self.text = ''
            elif self.cmd is Requests.STOP_RUN:
                self.factory.node.stop_run()
                #TODO Return something meaningful
                self.__answer("Meaningless")
            elif self.cmd is Requests.VERSION:
                self.__answer(self.factory.node.VERSION)
            elif self.cmd is Requests.RESULT_DIR:
                self.__answer(self.factory.node.get_result_dir())
            else:
                print "unknown cmd" + self.text
        else:
            # More data is coming
            if self.cmd is Requests.START_RUN:
                # Expect two lines: First is integer (scanrun.id), second
                # line is json ScanConfiguration.
                if self.nr_lines_received == 2:
                    self.factory.node.scanrunid = int(self.text)
                    self.text = ''
                else:
                    scan_config = json.loads(self.text, object_hook=mapwoc_from_json_hook)
                    if self.factory.node.start_job(scan_config):
                        self.__answer(json_trsuccess())
                    else:
                        self.factory.node.scanrunid = -1
                        self.__answer(json_trerror(500, "Node did not start the job"))
            elif self.cmd is Requests.CHECK_RESSOURCES:
                visitorconf = json.loads(self.text, object_hook=mapwoc_from_json_hook)
                ram_left = self.factory.node.job_feasible(visitorconf)

                if ram_left > 0:
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(101, "RAM: %d" % ram_left))
            else:
                print "Unexpected data."


    def connectionLost(self, reason):
        """Callback when connection lost."""
        pass

    def __answer(self, answer):
        """
        Write string answer to wire and close connection.
        @param answer: Answer to send as is.
        """

        self.transport.write(answer)
        self.transport.loseConnection()
       
    def __answer_json(self, answer):
        """
        Sends json encoded answer through the wire and closes the
        connection afterwards.
        @param answer: Answer to encode and send.
        """

        json_str = MapWOCJSONEncoder().encode(answer)
        self.__answer(json_str)


class NodeResponseFactory(Factory):
    """Factory producing NodeRequest-protocols."""

    protocol = NodeRequest
    def __init__(self, node):
        self.node = node

def setup_logging(quiet, verbose):
    """Setup logging."""

    mainlogger = logging.getLogger('nodectrl')
    nodelogger = logging.getLogger('node')
    serverlogger = logging.getLogger('nodectrlsrv')
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    if verbose:
        min_level = logging.DEBUG
    elif quiet:
        min_level = logging.WARNING
    else:
        min_level = logging.INFO

    for loggr in (mainlogger,serverlogger,nodelogger):
        lfh = logging.FileHandler('node.log')
        lsh = logging.StreamHandler()
        lfh.setFormatter(formatter)
        lsh.setFormatter(formatter)

        loggr.addHandler(lfh)
        loggr.addHandler(lsh)
        loggr.setLevel(min_level)


def main():
    """Entry point."""

    description = ('mapwoc-node listens to commands from a mapwoc-master.  '
                   'These commands typically result in start and stop of kvm '
                   'processes and analysis of network traffic created by those '
                   'virtual machines run by the kvm processes.')
    version = 'mapwoc-node (twisted wrapper) %s' % Node.VERSION

    combiopts = [
                ('v','verbose',"Be verbose and chatty on stdout and logfile.",
                 False,MapwocOptionParser.COMBIOPT_TYPE_MAKETRUE),
                ('q','quiet',"Be quiet, only log warnings and errors.",
                 False,MapwocOptionParser.COMBIOPT_TYPE_MAKETRUE),
                ('', 'ctrl_port',"Port to listen on for control commands.",
                 18158,MapwocOptionParser.COMBIOPT_TYPE_INT),
                ('', 'hcsdir',"Directory from which to read HC files.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'work_dir',"Directory in which to work and store temporary files.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'result_dir',"Directory in which to place results that are fetched from master.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ('', 'num_analyser',"Number of analyser processes to start in parallel.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_INT),
                ('', 'max_gb_ram',"Number of analyser processes to start in parallel.",
                 None,MapwocOptionParser.COMBIOPT_TYPE_FLOAT),
                ('', 'analyser',"The program to execute for analysis",
                 None,MapwocOptionParser.COMBIOPT_TYPE_STRING),
                ]
    parser = MapwocOptionParser(version, description,
                          "/etc/mapwoc/node/node.config", combiopts)

    (options, args) = parser.parse_args()
    if options.quiet and options.verbose:
        print "--verbose takes precedence over --quiet"

    setup_logging(options.quiet, options.verbose)
    logger = logging.getLogger('nodectrl')

    
    node = Node(options.hcsdir, options.work_dir, options.result_dir,
                options.num_analyser, options.max_gb_ram, options.analyser)
    factory = NodeResponseFactory(node)

    port = options.ctrl_port
    reactor.listenTCP(port, factory)

    node_poll_hcs = task.LoopingCall(node.poll_hcs)
    node_poll_hcs.start(10)

    try:
        logger.debug("Starting Node Twisted Reactor.")
        reactor.run()
    except:
        print "probably killed."
        #reactor.stop()
    print "Exiting."
    exit(0)

if __name__ == '__main__':
    main()

