import logging
import datetime
import time

from pylons import request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect

import formencode
from pylons.decorators import validate
from pylons.decorators.rest import restrict
from pylons.i18n.translation import _

from sqlalchemy import desc

import webhelpers.paginate as paginate

from mapwocgui.lib.base import BaseController, render

from mapwocgui.model import Session
from mapwoc.lib.result import Visit, Result, ResultType
from mapwoc.lib.scanconfiguration import ScanRun
from mapwoc.lib.url import URL
from paste.fileapp import FileApp
from os import path

import tempfile
import tarfile
import mapwoc.lib.model.result
import mapwocgui.model.masterproxy as master
import mapwocgui.lib.helpers as h

log = logging.getLogger(__name__)

class FilterForm(formencode.Schema):
    messages = {'badFrom' : _(u"Invalid date format in field from:"),
                'badTo' :   _(u"Invalid date format in field to:"),
                'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    searchmask = formencode.validators.String()

    def validate_python(self, values, c):
        if not values.get("datefilter"):
            return
        try:
            date_from = time.strptime(values['datefrom'], "%d.%m.%Y")
        except:
            h.flash(self.message("badFrom", c))
            raise formencode.validators.Invalid(
                      self.message("no_message", c), values, c)
        try:
            date_to = time.strptime(values['dateto'], "%d.%m.%Y")
        except:
            h.flash(self.message("badTo", c))
            raise formencode.validators.Invalid(
                      self.message("no_message", c), values, c)

class ResultController(BaseController):

    def index(self):
        visits = Session.query(Visit).order_by(desc(Visit.date))
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        c.resulttypes = Session.query(ResultType).all()
        c.dtnow = datetime.datetime.now()
        c.scanruns = Session.query(ScanRun).all()
        if c.scanruns:
            c.dtfirst = min([scan.starttime for scan in c.scanruns])
        if not c.dtfirst:
            c.dtfirst = c.dtnow
        return render('/result/index_result.mako')

    @validate(schema=FilterForm(), form='index')
    def filter(self):
        searchmask = self.form_result.get('searchmask')
        types = self.form_result.get('resultType')
        usedate = self.form_result.get('datefilter')
        query = Session.query(Visit)
        scan_id = self.form_result.get('scanrun')
        date_to = self.form_result.get('dateto')
        date_from = self.form_result.get('datefrom')
        if searchmask:
            query = query.filter(Visit.urls.any(
                URL.url.like("%%%s%%" % searchmask)))

        if len(types) and not "any" in types:
            for type in types:
                query = query.filter(
                            Visit.results.any(result_type_id = int(type)))

        if scan_id and scan_id != "any":
            query = query.filter_by(scanrun_id = scan_id)

        if usedate:
            datefrom = time.strptime(date_from, "%d.%m.%Y")
            dateto = time.strptime(date_to, "%d.%m.%Y")
            query = query.filter(Visit.date.between(datefrom, dateto))
            return datefrom
        visits = query.all()
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        c.filtered_view = True
        return render('/result/index_result.mako')

    def by_type(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter(
                Visit.results.any(result_type_id = int(type)))
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def by_scanrun(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter_by(scanrun_id=id)
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def by_url(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter(Visit.urls.any(id=id))
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def view_visit(self, id):
        c.visit = Session.query(Visit).filter_by(id=id).first()
        if c.visit is None:
            abort(404)
        c.scanrun = c.visit.scan_run
        c.sc = c.scanrun.scanconfiguration
        c.types_found = []
        # TODO: Localize
        c.firstDate = c.visit.results[0].date.strftime("%d.%m.%Y - %H:%M:%S")
        for result in c.visit.results:
            if not result.result_type in c.types_found:
                c.types_found.append(result.result_type)
        c.highest_severity = max([type.severity for type in c.types_found])
        c.hc = master.get_hc(c.visit.hcsid)

        return render('/result/view_visit.mako')

    def send_dump(self, id):
        """ Sends the results dump """
        # TODO gzip
        visit = Session.query(Visit).filter_by(id=id).first()
        if visit is None:
            return "Visit id not found"
        if not visit.filepath:
            return "Visit has no saved files"
        tgz = tarfile.open(visit.filepath)
        dumpname = ""
        for name in tgz.getnames():
            if name.endswith(".dump"):
                dumpname = name
                break
        else:
            return "Visit has no dump"
        tfile = tgz.extractfile(dumpname)
        user_filename = "%s_%s.dump" % (visit.hcsid, id)

        response.headers = [('Content-Disposition', 'attachment; filename=\"' + \
                        user_filename + '\"'),
                   ('Content-Type', 'text/plain'),
                   ('Content-Length', str(tfile.size))]
        def file_streamer():
            while True:
                block = tfile.read(4096)
                if not block:
                    break
                yield block
        response.app_iter = file_streamer()


      #  return tmpfile.read()

    def send_tarball(self, id):
        """ Sends the result tarball """
        visit = Session.query(Visit).filter_by(id=id).first()
        if visit is None:
            return "Visit id not found"

        if not visit.filepath:
            return "Visit has no saved files"
        user_filename = "Result_%s.tar.gz" % id
        file_size = path.getsize(visit.filepath)

        headers = [('Content-Disposition', 'attachment; filename=\"' + \
                        user_filename + '\"'),
                   ('Content-Type', 'text/plain'),
                   ('Content-Length', str(file_size))]

        fapp = FileApp(visit.filepath, headers=headers)
        return fapp(request.environ, self.start_response)

    def view_result(self, id):
        c.result = Session.query(Result).filter_by(id=id).first()
        if c.result is None:
            abort(404)
        c.visit =  c.result.visit
        c.hc = master.get_hc(c.visit.hcsid)
        c.sc = c.visit.scan_run
        return render('/result/view_result.mako')

    def view_dummy(self):
        return render('/result/view_result.mako')
