import logging

from pylons import request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect, redirect_to
from pylons.i18n.translation import _, ungettext

import formencode
from pylons.decorators import validate
from pylons.decorators.rest import restrict

import webhelpers.paginate as paginate

import mapwocgui.lib.helpers as h

from mapwocgui.lib.base import BaseController, render

from mapwocgui.model import Session
from mapwoc.lib.url import URLList, URL, create_URL_List
from mapwoc.lib.scanconfiguration import ScanConfiguration
from mapwoc.lib.scanconfiguration import TransportableSC
from mapwoc.lib.scanconfiguration import VisitorConfiguration
from mapwoc.lib.scanconfiguration import VisitorBundle
from mapwoc.lib.trivialresponse import TrivialResponse

import mapwocgui.model.masterproxy as master

from time import strftime

log = logging.getLogger(__name__)

# TODO initially, no URL Lists exist, thus no scan can be created! Fix that
# TODO move validators to lib

# TODO Validate that the QuickScan List is not used except for QuickScans.

class HCSelectionValidator(formencode.validators.FancyValidator):
    """ Validates the HoneyClient list. """

    messages = {'empty' : _(u"You must at least select one HoneyClient."),
                'negativeNumber' : _(u"Cannot start a negative number of HoneyClients."),
                'noNode' : _(u"One ore more HoneyClients do"
                              " not appear to be on any Node."),
                'resourceCheckFail' : _(u"This configuration did not pass"
                              " the ressource check on one or more nodes."),
                'no_message' : " "}

    def validate_python(self, values, c):
        atleastone = False
        visitors = {}

        hckeys = [key for key in values.keys() if key.startswith('hc')]
        for key in hckeys:
            if int(values[key]) < 0:
                h.flash(self.message('negativeNumber' ,c ))
                raise formencode.validators.Invalid(
                        self.message("no_message", c), values, c)
            if int(values[key]):
                atleastone = True
            if len(key.split(",")) < 2:
                h.flash(self.message("noNode", c))
                raise formencode.validators.Invalid(
                        self.message("no_message", c), values, c)

        if not atleastone:
            # Workaround to make the error message look consistent
            h.flash(self.message("empty", c))
            raise formencode.validators.Invalid(self.message("no_message", c),
                    values, c)

class RessourceCheckValidator(formencode.validators.FancyValidator):
    """
    Ask the master to asks its node whether they have the ressources to start
    this scan.
    """

    messages = {
                'no_message' : " ",
                'not_enough_ressources' : _(u"Ressource Check Failed:"
                     " One or more Nodes do not have sufficient ressources.")
               }

    def validate_python(self, values, c):
        sc = scanconfig_from_form(values)

        # The URL-Lists from this scanconfig are created from a form
        # and not read from database, thus not connected to db-Session.
        # Establish this connection now. 
        Session.add(sc)
        dburllists = []
        for ul in sc.urlcollection:
            dbul = Session.query(URLList).filter_by(name=ul.name).first()
            dburllists.append(dbul)
        sc.urlcollection = dburllists

        tr = master.check_scan(TransportableSC(sc))
        Session.rollback()
        if not tr.error:
            pass
        else:
            h.flash(self.message('not_enough_ressources', c) + tr.message)
            raise formencode.validators.Invalid(self.message("no_message", c),
                                                values, c)

class ScanForm(formencode.Schema):
    """Schema for the form to create a Scan."""

    chained_validators = [HCSelectionValidator, RessourceCheckValidator]
    allow_extra_fields = True
    filter_extra_fields = False
    name = formencode.validators.PlainText(not_emtpy=True)
    description = formencode.validators.String()
    comment = formencode.validators.String(not_empty=True)
    active_window = formencode.validators.Int(not_empty=True)
    urllist = formencode.validators.NotEmpty()
    messages = {'no_message' : " "}

    def validate_python(self, values, c):
        if values["name"] == "QuickScan":
            h.flash(_(u"Name QuickScan is reserved"))
            raise formencode.validators.Invalid(self.message("no_message", c),
                   values, c)


class QuickformURLValidator(formencode.validators.FancyValidator):
    messages = {'no_message' : " "}

    def validate_python(self, values, c):
        urllist,invalids = create_URL_List("tmp",
                                    values, None)
        for invalid in invalids:
            h.flash(_(u"Invalid URL: %s") % invalid)
        if invalids:
            raise formencode.validators.Invalid(self.message("no_message", c),
                   values, c)
        return values

class QuickForm(formencode.Schema):
    messages = {'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    quickurl = QuickformURLValidator()

    # TODO Validate VM.
    def validate_python(self, values, c):
        check_running_scan(self, values, c)

def check_running_scan(self, values, c):
    c.current_scan_name = master.get_state_set().get('scanname')
    if c.current_scan_name:
        h.flash(
            _(u"Quickscan can not be started while %s is still running.") % \
                c.current_scan_name)
        raise formencode.validators.Invalid(self.message("no_message", c),
               values, c)

def scanconfig_from_form(form_result):
    sc = ScanConfiguration(form_result['name'])
    sc.comment = form_result['comment']
    sc.description = form_result['description']
    vc = VisitorConfiguration()
    vc.active_window = form_result['active_window']
    if form_result.get('fastscan'):
        vc.num_urls = form_result['num_urls']
    sc.visitor_configuration = vc
    for selected in form_result['urllist']:
        sc.urlcollection += [Session.query(URLList).get(int(selected))]

    for k in form_result:
        if k.startswith("hc") and int(form_result[k]) != 0:
            node = k.split(",")[1]
            vc.visitors.append(VisitorBundle(k[2:].split(",")[0],
                int(form_result[k]), node))

    return sc

class ScanController(BaseController):

    def index(self):
        scanconfigurations = Session.query(ScanConfiguration)
        c.scanconfigurations_paginator = paginate.Page(scanconfigurations,
                                     page=int(request.params.get('page', 1)),
                                     items_per_page=20,)
        c.hcs = master.get_hcs()
        c.available_vms = []
        for hc in c.hcs:
            for node in hc.on_nodes:
                c.available_vms.append(hc.hcsid + ", " + node)
        c.current_scan_name = master.get_state_set().get('scanname')
        return render('/scan/index_scan.mako')

    def view(self, id):
        c.scanconfig = Session.query(ScanConfiguration).get(int(id))
        hcs = master.get_hcs()
        # TODO use list comprehensions
        c.hcs = {}
        for hc in hcs:
            c.hcs[hc.hcsid] = hc
        return render('/scan/view_scan.mako')

    def create(self):
        all_urllists = Session.query(URLList).filter_by(deleted=False).all() 
        c.available_urllists = [(u.id, u.name) for u in all_urllists]

        if not c.available_urllists:
            h.flash(_(u"No Urllists avaiable to create scan"))

        c.hcs = master.get_hcs()
        return render('scan/create_scan.mako')

    # TODO Should be post_only=True?
    @validate(schema=ScanForm(), form='create', post_only=False, on_get=True)
    def new(self):
        # Create the ScanConfig and register it in database.
        sc = scanconfig_from_form(self.form_result)

        Session.add(sc)
        Session.commit()
        
        redirect_to(controller='scan', action='view', id=sc.id)

    def delete(self, id):
        scanc = Session.query(ScanConfiguration).filter_by(id=id).first()
        if scanc is None:
            abort(404)
        Session.delete(scanc)
        Session.commit()

        redirect_to(controller='scan', action='index')

    def list(self):
        c.scanconfigurations = Session.query(ScanConfiguration).all()
        return render('/scan/list_scans.mako')

    def start(self, id):
        scanconfig = Session.query(ScanConfiguration).get(int(id))

        response = master.start_scan(TransportableSC(scanconfig))

        # TODO straighten this up
        if response.error == 0:
            redirect_to(controller='state',action='index')
        else:
            return "Trivial response of master indicated error."

    def stop(self):
        """ Stops any running scan. """
        master.stop_scan(None)
        # TODO fetch feedback
        redirect_to(controller='state',action='index')

    @validate(schema=QuickForm(), form='index', post_only=False, on_get=True)
    def quickscan(self):
        qsc = Session.query(ScanConfiguration).filter_by(name="QuickScan").first()
        qul = Session.query(URLList).filter_by(name="QuickScan").first()

        all_urllists = Session.query(URLList).all()

        if not qsc:
            qsc = ScanConfiguration( "QuickScan" )

        qsc.comment = _(u"Last changed: %s") % strftime("%Y-%m-%d-%H-%M-%S")
        vc = VisitorConfiguration()
        vc.visitors.append(VisitorBundle(self.form_result['vm'].split(", ")[0],
                1, self.form_result['vm'].split(", ")[1]))

        if qul:
            qul.urls = [URL(self.form_result['quickurl'])]
            qul.comment = qsc.comment
        else:
            urllist, invalids = create_URL_List("QuickScan",
                                    self.form_result['quickurl'],
                                    qsc.comment)
            qul = urllist

        vc.active_window = 60
        qsc.visitor_configuration = vc

        qsc.urlcollection = [qul]

        Session.add(qul)
        Session.add(qsc)
        Session.commit()

        return self.start(qsc.id)

